import abc

import numpy
from sklearn import linear_model
from sklearn import tree

from ..utils import constants
from ..utils import helpers

logger = helpers.make_logger(__name__)


class MetaMLModel(abc.ABC):

    def __init__(self, model_class, **kwargs):
        """ Creates a model and initializes it with the given key-word arguments.
        """
        self.model = model_class(**kwargs)

    @property
    @abc.abstractmethod
    def name(self) -> str:
        pass

    @abc.abstractmethod
    def fit(self, *args, **kwargs) -> 'MetaMLModel':
        """ Fits the model.
        """
        pass

    @abc.abstractmethod
    def predict(self, ratios: numpy.ndarray) -> float:
        """ Predicts using the underlying model.
        """
        pass

    @abc.abstractmethod
    def extract_python(self, *args, **kwargs) -> tuple[list[str], str]:
        """ Extracts the scoring function as a string which can be written to disk.

        The scoring function should have the following signature:

        ```
        helpful_function_name(ratios: numpy.ndarray) -> float
        ```

        where `ratios` is a 1d array of floats with 6 elements. These ratios are
        `cluster_ratios` from Manifold.

        Returns:
            A 2-tuple whose items are:
            - a list of strings which will be written as the import lines.
            - a string containing the code for the scoring function in Python.
        """
        pass


class MetaDT(MetaMLModel):

    def __init__(self):
        super().__init__(tree.DecisionTreeRegressor, max_depth=3)

    @property
    def name(self) -> str:
        return 'dt'

    def fit(self, data: numpy.ndarray, scores: numpy.ndarray) -> 'MetaDT':
        logger.info(f'Fitting meta-model {self.name} on data with shape {data.shape} ...')
        self.model = self.model.fit(data, scores)
        return self

    def predict(self, ratios: numpy.ndarray) -> float:
        return self.model.predict(ratios)

    def extract_python(self, metric: str, method: str) -> tuple[list[str], str]:
        # noinspection PyProtectedMember
        undefined_feature = tree._tree.TREE_UNDEFINED

        feature_names = [
            constants.RATIO_NAMES[i] if i != undefined_feature else "undefined!"
            for i in self.model.tree_.feature
        ]

        # start function definition
        function_name = f'{self.name}_{metric}_{method}'
        code_lines: list[str] = [
            f'def {function_name}(ratios: numpy.ndarray) -> float:',
            f'    {", ".join(constants.RATIO_NAMES)} = tuple(ratios)',
        ]

        def extract_lines(node_index: int, indent: str):
            if self.model.tree_.feature[node_index] != undefined_feature:  # internal node

                feature: str = feature_names[node_index]
                threshold: float = self.model.tree_.threshold[node_index]
                left_index: int = self.model.tree_.children_left[node_index]
                right_index: int = self.model.tree_.children_right[node_index]

                # if block
                code_lines.append(f'{indent}if {feature} <= {threshold:.6e}:')
                extract_lines(left_index, indent + '    ')

                # else block
                code_lines.append(f'{indent}else:')
                extract_lines(right_index, indent + '    ')

            else:  # leaf node
                # return
                value: float = self.model.tree_.value[node_index][0][0]
                code_lines.append(f'{indent}return {value:.6e}')

        extract_lines(0, '    ')
        code = '\n'.join(code_lines)

        imports = ['import numpy']

        return imports, code


class MetaLR(MetaMLModel):

    def __init__(self):
        super().__init__(linear_model.LinearRegression, fit_intercept=False)

    @property
    def name(self) -> str:
        return 'lr'

    def fit(self, data: numpy.ndarray, scores: numpy.ndarray) -> 'MetaLR':
        logger.info(f'Fitting meta-model {self.name} on data with shape {data.shape} ...')
        self.model = self.model.fit(data, scores)
        return self

    def predict(self, ratios: numpy.ndarray) -> float:
        return self.model.predict(ratios)

    def extract_python(self, metric: str, method: str) -> tuple[list[str], str]:
        imports = ['import numpy']

        function_name = f'{self.name}_{metric}_{method}'
        coefficients = [f'{float(c):.6e}' for c in self.model.coef_]
        code = '\n'.join([
            f'def {function_name}(ratios: numpy.ndarray) -> float:',
            f'    return float(numpy.dot(numpy.asarray(',
            f'        a=[{", ".join(coefficients)}],',
            f'        dtype=float,',
            f'    ), ratios))'
        ])

        return imports, code


__all__ = [
    'MetaMLModel',
    'MetaDT',
    'MetaLR',
]



from . import cluster as cluster
from . import cluster_criteria as cluster_criteria
from . import dataset as dataset
from . import graph as graph
from . import graph_criteria as graph_criteria
from . import metric as metric
from . import space as space
from .cluster import Cluster as Cluster
from .cluster_criteria import ClusterCriterion as ClusterCriterion
from .dataset import Dataset as Dataset
from .graph import Edge as Edge
from .graph import Graph as Graph
from .graph_criteria import GraphCriterion as GraphCriterion
from .metric import Metric as Metric
from .space import Space as Space



import logging
import os

LOG_LEVEL = getattr(logging, os.environ.get('CLAM_LOG', 'INFO'))
EMA_SMOOTHING = float(os.environ.get('CLAM_EMA_SMOOTHING', '2'))
EMA_PERIOD = float(os.environ.get('CLAM_EMA_PERIOD', '10'))
EMA_ALPHA = EMA_SMOOTHING / (1 + EMA_PERIOD)

SUBSAMPLE_LIMIT = 100
BATCH_SIZE = 10_000
EPSILON = 1e-6

Ratios = tuple[float, float, float, float, float, float]
RATIO_NAMES = ['cardinality', 'radius', 'lfd']
RATIO_NAMES.extend([f'{name}_ema' for name in RATIO_NAMES])


class Unset:
    """ This is a hack around type-hinting when a value cannot be set in the
    __init__ method for a class.

    https://peps.python.org/pep-0661/

    Usage:

    ```python

    class MyClass:
        def __init__(self, *args, **kwargs):
            ...
            self.__value: typing.Union[ValueType, Unset] = UNSET

        def value_setter(self, *args, **kwargs):
            ...
            self.__value = something
            return

        @property
        def value(self) -> ValueType:
            if self.__value is UNSET:
                raise ValueError(f'Please call `value_setter` on the object before using this property.')
            return self.__value
    ```
    """
    __unset = None

    def __new__(cls):
        if cls.__unset is None:
            cls.__unset = super().__new__(cls)
        return cls.__unset


UNSET = Unset()



import abc
import typing

import numpy

from .. import core
from ..utils import constants
from ..utils import helpers

logger = helpers.make_logger(__name__)

ClusterScores = dict[core.Cluster, float]
InstanceScores = dict[int, float]

# TODO: Look at DSD (diffusion-state-distance) as a possible individual-method.
# TODO: Specialize `should_be_fast` for each algorithm based on complexity analysis and graph properties.


class GraphScorer(abc.ABC):

    def __call__(self, g: core.Graph) -> tuple[ClusterScores, numpy.ndarray]:
        """ Use this scorer to generate normalized anomaly scores.
        """
        logger.info(f'Running method {self.name} on a graph with {g.vertex_cardinality} clusters.')

        cluster_scores = self.score_graph(g)
        if self.normalize_on_clusters:
            cluster_scores = self.__normalize_scores(cluster_scores)

        instance_scores = self.inherit_scores(cluster_scores)
        if not self.normalize_on_clusters:
            instance_scores = self.__normalize_scores(instance_scores)

        scores_array = self.ordered_scores(g, instance_scores)
        return cluster_scores, scores_array

    @abc.abstractmethod
    def __hash__(self):
        """ A way to uniquely identify each `GraphScorer` object.
        """
        pass

    @property
    @abc.abstractmethod
    def name(self) -> str:
        """ Full name of the algorithm.
        """
        pass

    @property
    @abc.abstractmethod
    def short_name(self) -> str:
        """ Abbreviated name of the algorithm.
        """
        pass

    @property
    @abc.abstractmethod
    def normalize_on_clusters(self) -> bool:
        """ Whether to normalize scores for clusters or for instances.
        """
        pass

    @property
    @abc.abstractmethod
    def normalization_mode(self) -> helpers.NormalizationMode:
        """ What normalization method to use.
        """
        pass

    @abc.abstractmethod
    def score_graph(self, g: core.Graph) -> ClusterScores:
        """ The method with which to assign anomaly scores to `Clusters` in the
        `Graph`. This should return a dictionary whose keys are `Clusters` in
        the `Graph` and whose values are the pre-normalization anomaly scores.
        """
        pass

    @abc.abstractmethod
    def should_be_fast(self, g: core.Graph) -> bool:
        """ Whether this algorithm is expected to run in a reasonably short
        time on the given `Graph`. This method should make a quick estimate.
        """
        pass

    def __normalize_scores(
            self,
            scores: typing.Union[ClusterScores, InstanceScores],
    ) -> typing.Union[ClusterScores, InstanceScores]:
        [keys, scores] = list(zip(*scores.items()))
        scores = helpers.normalize(numpy.asarray(scores, dtype=numpy.float32), mode=self.normalization_mode)
        return {c: s for c, s in zip(keys, map(float, scores))}

    @staticmethod
    def inherit_scores(scores: ClusterScores) -> InstanceScores:
        return {i: s for c, s in scores.items() for i in c.indices}

    @staticmethod
    def ordered_scores(g: core.Graph, scores: InstanceScores) -> numpy.ndarray:
        [indices, scores] = list(zip(*sorted(scores.items())))
        assert set(indices) == {i for c in g.clusters for i in c.indices}
        return numpy.asarray(scores, dtype=numpy.float32)


class ClusterCardinality(GraphScorer):

    def __hash__(self):
        return hash(self.name)

    @property
    def name(self) -> str:
        return 'cluster_cardinality'

    @property
    def short_name(self) -> str:
        return 'cc'

    @property
    def normalize_on_clusters(self) -> bool:
        return True

    @property
    def normalization_mode(self) -> helpers.NormalizationMode:
        return 'gaussian'

    def should_be_fast(self, _) -> bool:
        return True

    def score_graph(self, g: core.Graph) -> ClusterScores:
        return {c: -c.cardinality for c in g.clusters}


class ComponentCardinality(GraphScorer):

    def __hash__(self):
        return hash(self.name)

    @property
    def name(self) -> str:
        return 'component_cardinality'

    @property
    def short_name(self) -> str:
        return 'sc'

    @property
    def normalize_on_clusters(self) -> bool:
        return True

    @property
    def normalization_mode(self) -> helpers.NormalizationMode:
        return 'gaussian'

    def should_be_fast(self, _) -> bool:
        return True

    def score_graph(self, g: core.Graph) -> ClusterScores:
        return {
            c: -component.vertex_cardinality
            for component in g.components
            for c in component.clusters
        }


class VertexDegree(GraphScorer):

    def __hash__(self):
        return hash(self.name)

    @property
    def name(self) -> str:
        return 'vertex_degree'

    @property
    def short_name(self) -> str:
        return 'vd'

    @property
    def normalize_on_clusters(self) -> bool:
        return True

    @property
    def normalization_mode(self) -> helpers.NormalizationMode:
        return 'gaussian'

    def should_be_fast(self, _) -> bool:
        return True

    def score_graph(self, g: core.Graph) -> ClusterScores:
        return {c: -g.vertex_degree(c) for c in g.clusters}


class ParentCardinality(GraphScorer):

    def __init__(self, depth_weight: typing.Callable[[int], float]):
        self.weight = depth_weight

    def __hash__(self):
        return hash(self.name)

    @property
    def name(self) -> str:
        return 'parent_cardinality'

    @property
    def short_name(self) -> str:
        return 'pc'

    @property
    def normalize_on_clusters(self) -> bool:
        return True

    @property
    def normalization_mode(self) -> helpers.NormalizationMode:
        return 'gaussian'

    def should_be_fast(self, _) -> bool:
        return True

    def score_graph(self, g: core.Graph) -> ClusterScores:

        scores = {c: 0 for c in g.clusters}

        for c in g.clusters:
            ancestry = c.ancestry
            for i in range(1, len(ancestry)):
                scores[c] += (self.weight(i) * ancestry[i - 1].cardinality / ancestry[i].cardinality)

        return {c: scores[c] for c in g.clusters}


class GraphNeighborhood(GraphScorer):

    def __init__(self, eccentricity_fraction: float):
        if not (0 < eccentricity_fraction <= 1):
            raise ValueError(f'eccentricity fraction must be in the (0, 1] range. Got {eccentricity_fraction:.2e} instead.')

        self.eccentricity_fraction = eccentricity_fraction

    def __hash__(self):
        return hash(self.name)

    @property
    def name(self) -> str:
        return 'graph_neighborhood'

    @property
    def short_name(self) -> str:
        return 'gn'

    @property
    def normalize_on_clusters(self) -> bool:
        return True

    @property
    def normalization_mode(self) -> helpers.NormalizationMode:
        return 'gaussian'

    def should_be_fast(self, g: core.Graph) -> bool:
        return g.vertex_cardinality < 256

    def num_steps(self, g: core.Graph, c: core.Cluster) -> int:
        return int(g.eccentricity(c) * self.eccentricity_fraction) + 1

    def score_graph(self, g: core.Graph) -> ClusterScores:

        step_sizes: dict[core.Cluster, list[int]] = {
            c: g.frontier_sizes(c)[:self.num_steps(g, c)]
            for c in g.clusters
        }
        scores = {c: -float(sum(sizes)) for c, sizes in step_sizes.items()}

        return scores


class StationaryProbabilities(GraphScorer):

    def __init__(self, steps: int):
        self.steps = steps

    def __hash__(self):
        return hash(self.name)

    @property
    def name(self) -> str:
        return 'stationary_probabilities'

    @property
    def short_name(self) -> str:
        return 'sp'

    @property
    def normalize_on_clusters(self) -> bool:
        return True

    @property
    def normalization_mode(self) -> helpers.NormalizationMode:
        return 'gaussian'

    def should_be_fast(self, g: core.Graph) -> bool:
        return True

    def score_graph(self, g: core.Graph) -> ClusterScores:

        scores = dict()

        for component in g.components:

            if component.vertex_cardinality > 1:
                clusters, matrix = component.distance_matrix

                # TODO: Figure out how to go until convergence.
                sums = numpy.sum(matrix, axis=1)[:, None] + constants.EPSILON
                matrix = matrix / sums
                for _ in range(self.steps):
                    matrix = numpy.linalg.matrix_power(matrix, 2)

                scores.update({c: s for c, s in zip(clusters, numpy.sum(matrix, axis=0))})

            else:  # Component with only one Cluster means that the Cluster is anomalous.
                scores.update({c: 0 for c in component.clusters})

        scores = {c: -s for c, s in scores.items()}

        return scores


__all__ = [
    'ClusterScores',
    'InstanceScores',
    'GraphScorer',
    'ClusterCardinality',
    'ComponentCardinality',
    'VertexDegree',
    'ParentCardinality',
    'GraphNeighborhood',
    'StationaryProbabilities',
]



from . import anomaly_detection as anomaly_detection
from . import classification as classification
from . import core as core
from . import search as search
from . import utils as utils
from .core import *



from . import anomaly_dataset as anomaly_dataset
from . import anomaly_space as anomaly_space
from . import chaoda as chaoda
from . import graph_scorers as graph_scorers
from . import meta_ml as meta_ml
from . import pretrained_models as pretrained_models
from . import training as training
from .anomaly_dataset import AnomalyDataset as AnomalyDataset
from .anomaly_space import AnomalySpace as AnomalySpace
from .chaoda import CHAODA as CHAODA



import abc
import heapq
import typing

import numpy

from . import cluster
from ..utils import helpers

logger = helpers.make_logger(__name__)


class GraphCriterion(abc.ABC):
    """ A `GraphCriterion` can be called on the `root` cluster of a tree to
    select a set of clusters from the tree. These clusters can be used to
    create a `Graph`. Subclasses much implement the `select` method. The
    abstract `GraphCriterion` will verify that the set of selected clusters
    obey the graph invariant. See the documentation for `Graph` for details on
    the graph invariant.
    """

    @property
    @abc.abstractmethod
    def name(self) -> str:
        pass

    @abc.abstractmethod
    def select(self, root: cluster.Cluster) -> set[cluster.Cluster]:
        pass

    @staticmethod
    def assert_invariant(root: cluster.Cluster, selected: set[cluster.Cluster]):
        for c in selected:
            if any((c.is_ancestor_of(other) for other in selected)):
                raise ValueError(f'A cluster and its ancestor were both selected.')

        indices = {i for c in selected for i in c.indices}

        if len(indices) != root.cardinality:
            raise ValueError(
                f'There was a mis-match in the number of instances that were '
                f'selected. The selected clusters have {len(indices)} instance '
                f'but the root has {root.cardinality}.'
            )

        return

    def __call__(self, root: cluster.Cluster) -> set[cluster.Cluster]:
        selected = self.select(root)
        self.assert_invariant(root, selected)
        return selected


class Layer(GraphCriterion):
    """ Selects the layer at the specified depth, -1 means go to leaves.
    """
    def __init__(self, depth: int):
        if depth < -1:
            raise ValueError(f'expected a \'-1\' or a non-negative depth. got: {depth}')
        self.depth = depth

    @property
    def name(self) -> str:
        return f'Layer_{self.depth}'

    def select(self, root: cluster.Cluster) -> set[cluster.Cluster]:

        if self.depth == -1:
            return {c for layer in root.subtree for c in layer if c.is_leaf}

        else:
            selected = {c for layer in root.subtree[:self.depth] for c in layer if c.is_leaf}
            selected.update(root.subtree[self.depth])

            return selected


class PropertyThreshold(GraphCriterion):
    """ Selects clusters when the given property crosses the given
    percentile.
    """

    def __init__(
            self,
            value: typing.Literal['cardinality', 'radius', 'lfd'],
            percentile: float,
            mode: typing.Literal['above', 'below'],
    ):
        """
        During a BFT of the tree, this keeps track of the given cluster
        property. A cluster is selected if any of the following is true:
            - it is a leaf, or
            - it qualifies by `mode`, or
            - both children qualify by `mode`.

        If `mode` is 'above', a cluster qualifies if the property goes above the
        `percentile`. If `mode` is 'below', a cluster qualifies if the property
        goes below the `percentile`.

        Args:
            value: The cluster property to keep track of. Must be one of
             'cardinality', 'radius', or 'lfd'.
            percentile: the percentile, in the (0, 100) range, to be crossed.
            mode: select a cluster when `value` is 'above' or 'below' the
             `percentile`.
        """
        if 0. < percentile < 100.:
            self.percentile: float = percentile
        else:
            raise ValueError(f'percentile must be in the (0, 100) range. Got {percentile:.2f} instead.')

        self.value = value

        if mode == 'above':
            self.qualifies = lambda c, v: getattr(c, self.value) > v
        elif mode == 'below':
            self.qualifies = lambda c, v: getattr(c, self.value) < v
        else:
            raise ValueError(f'mode must be \'above\' or \'below\'. Got {mode} instead.')

    @property
    def name(self) -> str:
        return f'PropertyThreshold_{self.value}_{self.name}_{self.percentile:.2f}'

    def select(self, root: cluster.Cluster) -> set[cluster.Cluster]:
        threshold = float(numpy.percentile(
            [getattr(c, self.value) for layer in root.subtree for c in layer if c.cardinality > 1],
            self.percentile,
        ))
        selected: set[cluster.Cluster] = set()
        frontier: set[cluster.Cluster] = {root}
        while frontier:
            c = frontier.pop()
            if (
                    c.is_leaf  # select the leaves
                    or self.qualifies(c, threshold)  # if cluster qualifies
                    or (self.qualifies(c.left_child, threshold) and self.qualifies(c.right_child, threshold))
            ):
                selected.add(c)
            else:  # add children to frontier
                frontier.update(c.children)
        return selected


class MetaMLSelect(GraphCriterion):
    """ Uses the scoring function from a trained meta-ml model to select
    clusters.
    """

    def __init__(self, scorer: typing.Callable[[numpy.ndarray], float], name: typing.Optional[str] = None, min_depth: int = 4):
        """
        Args:
            scorer: A function that takes the ratios of a cluster and returns
             its predicted score. Higher scoring clusters are considered
             better than lower scoring clusters.
            min_depth: The minimum depth in the tree for a cluster to be
             selected.
        """
        if min_depth < 1:
            raise ValueError(f'min-depth must be a positive integer.')

        self.__name = scorer.__name__ if name is None else name
        self.scorer = lambda c: -scorer(numpy.asarray(c.ratios, dtype=numpy.float32))
        self.min_depth = min_depth

    @property
    def name(self) -> str:
        return self.__name

    def select(self, root: cluster.Cluster) -> set[cluster.Cluster]:

        tree = [c for layer in root.subtree for c in layer]

        candidate_clusters = [c for c in tree if c.depth >= self.min_depth]
        normalized_scores = list(map(float, helpers.normalize(
            numpy.asarray(list(map(self.scorer, candidate_clusters))),
            mode='gaussian',
        )))

        heap = list(zip(normalized_scores, candidate_clusters))
        heapq.heapify(heap)

        selected = {c for c in tree if c.depth < self.min_depth and c.is_leaf}
        selected_indices = {i for c in selected for i in c.indices}

        while len(heap) > 0:
            _, c = heapq.heappop(heap)

            if len(selected_indices.intersection(set(c.indices))) > 0:
                continue
            else:
                selected.add(c)
                selected_indices.update(set(c.indices))

        return selected


__all__ = [
    'GraphCriterion',
    'Layer',
    'PropertyThreshold',
    'MetaMLSelect',
]



import operator
import typing

import numpy

from .. import core
from ..anomaly_detection import CHAODA
from ..utils import helpers

logger = helpers.make_logger(__name__)


class Classifier:

    def __init__(
            self,
            labels: numpy.ndarray,
            metric_spaces: list[core.Space],
            **kwargs,
    ):
        """ Creates and initializes a CLAM Classifier.

        Args:
            metric_spaces: See `CHAODA`.
            labels: 1d array of labels. dtype must be numpy.uint.
            kwargs: These are the same as the kwargs for `CHAODA`.
        """
        if not labels.dtype == numpy.uint:
            raise ValueError(f'labels must have dtype {numpy.uint}. Got {labels.dtype} instead.')

        self.__metric_spaces = metric_spaces
        self.__labels = list(map(int, labels))
        self.__unique_labels = list(set(self.__labels))
        self.__kwargs = kwargs
        self.__bowls: typing.Dict[int, CHAODA] = dict()

    @property
    def labels(self) -> list[int]:
        return self.__labels

    @property
    def unique_labels(self) -> list[int]:
        return self.__unique_labels

    def build(self) -> 'Classifier':
        """ Fits the Classifier to the data and returns the fitted object.
        """
        for label in self.__unique_labels:
            logger.info(f'Fitting CHAODA object for label {label} ...')

            indices = [i for i, l in enumerate(self.__labels) if l == label]
            metric_spaces = [s.subset(indices, f'{s.data.name}__{label}') for s in self.__metric_spaces]

            self.__bowls[label] = CHAODA(metric_spaces, **self.__kwargs).build()

        return self

    def rank_single(self, query) -> list[tuple[int, float]]:
        """ Predicts the class rankings for a single query. Lower scores are
        better.
        """
        label_scores = list()
        for label, bowl in self.__bowls.items():
            score = bowl.predict_single(query)
            label_scores.append((label, score))

        return label_scores

    def rank(self, queries: core.Dataset) -> list[list[tuple[int, float]]]:
        """ Predicts the class rankings for a set of queries. Lower scores are
        better.
        """
        label_scores = list()
        for i in range(queries.cardinality):
            logger.info(f'Predicting class for query {i} ...')
            label_scores.append(self.rank_single(queries[i]))
        return label_scores

    def predict_single(self, query) -> tuple[int, float]:
        """ Predicts the label for a single query.
        """
        label_scores = self.rank_single(query)
        best_label, best_score = min(label_scores, key=operator.itemgetter(1))
        return best_label, best_score

    def predict(self, queries: core.Dataset) -> tuple[list[int], list[float]]:
        """ Predicts the label for a set of queries.
        """
        label_scores = list()
        for i in range(queries.cardinality):
            logger.info(f'Predicting class for query {i + 1}/{queries.cardinality} ...')
            label_scores.append(self.predict_single(queries[i]))
        [labels, scores] = list(zip(*label_scores))
        return labels, scores


__all__ = [
    'Classifier',
]



import abc

from . import cluster
from ..utils import helpers

logger = helpers.make_logger(__name__)


class ClusterCriterion(abc.ABC):
    """ A rule to decide when a cluster can be partitioned. Subclasses must
    implement the `__call__` method to take a single `Cluster` and return a
    `bool` to indicate whether that cluster can be partitioned.

    If multiple criteria are used during `partition`, they must all return
    `True` to allow the cluster to be partitioned.
    """

    @abc.abstractmethod
    def __call__(self, c: 'cluster.Cluster') -> bool:
        pass


class MaxDepth(ClusterCriterion):
    """ Clusters with `depth` less than `max_depth` may be partitioned.
    """

    def __init__(self, max_depth: int):
        self.max_depth = max_depth

    def __call__(self, c: 'cluster.Cluster') -> bool:
        return c.depth < self.max_depth


class MinPoints(ClusterCriterion):
    """ Clusters with `cardinality` greater than `min_points` may be
    partitioned.
    """

    def __init__(self, min_points: int):
        self.min_points = min_points

    def __call__(self, c: 'cluster.Cluster') -> bool:
        return c.cardinality > self.min_points


class NotSingleton(ClusterCriterion):
    """ Clusters can be partitioned so long as they are not singletons.
    """

    def __call__(self, c: 'cluster.Cluster') -> bool:
        return not c.is_singleton


__all__ = [
    'ClusterCriterion',
    'MaxDepth',
    'MinPoints',
]
